#include <cmath>

enum Increment{
	sedgewick, //Sedgewick increments
	two_increment, //two elements:  1 and 1,72 x  (size) ^(1/3)
	product_of_three //shell_3voudsort, met een list with k0 = 1,  ki+1 = 3ki+1. 	
};

template<class T>
class ShellSort  : public virtual SortAlgorithm<T> {
	
	private: T* tab;
	private: int size;
	private: int* increments_table;
	private: int increments_size;
	private: Increment increment;
		
	private: void initializeSedgewick(){
		int sedgewick_inc[]={1,5,19,41,109,209,505,929,2161,
		3905,8929,16001,36289,64769,146305,260609,587521,1045505,2354689,
		4188161,9427969,16764929,37730305,67084289,150958081,268386305,
		603906049,1073643521};
		increments_size=0;
		int element=sedgewick_inc[increments_size];
		while(element< (int) (size/3)){
			element=sedgewick_inc[increments_size++];
		}
		increments_table = new int [increments_size];
		for(int i=0;i<increments_size;i++){
			increments_table[i]=sedgewick_inc[i];
		}
	}
	
	
	private: void initizalizeTwoIncrement(){
		double second_element= 1.72 * pow((double)size,(double)(1/3));
		increments_size=2;
		increments_table = new int [increments_size];
		increments_table[0]=1;
		increments_table[1] = (int) second_element;		
	}
	
	private: void initializeProductOfthree(){
		increments_size=1;
		int k = 1;
		while(k< (int) (size/3)){
			k=3*k+1;
			increments_size++;
		}
		increments_table = new int [increments_size];
		int i=0;
		k=1;		
		while(k<(int) (size/3)){
			increments_table[i]=k;
			k=3*k+1;
			i++;
		}
	}
	
	private: void initializeIncrements(){
		switch(increment){
			case sedgewick:
				initializeSedgewick();
				break;
			case two_increment:
				initizalizeTwoIncrement();
				break;
			case product_of_three:
				initializeProductOfthree();
				break;
		}
	}
	
	public: void sort(){
		int increment_index=increments_size-1;
		//for each increment:
		while(increment_index>=0){
        int k = increments_table[increment_index];
        InsertionSort<T>::insertionSort(tab,size,k,k);
        increment_index--;	
    }
	}
	
	public: ShellSort(T* table,int table_size,Increment  incr){
		size = table_size;
		tab=table;
		increment = incr;
		initializeIncrements();		
	}
	
	public: ~ShellSort(){
		delete[] increments_table;
	}
	
	private: string incrementName(){
		string name;
		switch(increment){
			case sedgewick:
				name="Sedgewick increments";
				break;
			case two_increment:
				name="Two Increment";
				break;
			case product_of_three:
				name="Product of three increment";
				break;
		}
		return name;
	}
	
	public: string getName(){
		return "Shell Sort: " + incrementName()  ;
	}
	
};







