#include <iostream>
#include <iomanip>
#include <string>

#include "SortAlgorithm.h"
#include "SelectionSort.h"
#include "InsertionSort.h"
#include "QuickSort.h"
#include "HeapSort.h"
#include "ShellSort.h"
#include "Stopwatch.h"
#include "QuickerSort.h"
#include "SixSort.h"
#include "CountingSort.h"
#include "MergeSort.h"
#include "QuickSmartPivotSort.h"

using namespace std;

enum Algorithm {
	selection_sort,
	insertion_sort,
	heap_sort,
	quick_sort,
	shell_sort_sedgewick,
	shell_sort_two_increment,
	shell_sort_product_of_three,
  quicker_sort,
  six_sort,
  counting_sort,
  radix_sort,
  merge_sort,
  quick_smart_pivot_sort
};

template<class T>
class Sorter{
	Algorithm algorithm;
	public: Sorter(Algorithm algo){
		algorithm = algo;
	}
	
	//reverses a table;
	private: void reverse(T* table, int size){
		for (int i=0; i<size/2;i++){
			T current=table[i];
			table[i]=table[size-i-1];  
			table[size-i-1]=current;
		}	
	}
  
  private: void shuffle(T* table,int size){
    for (int i=0; i<size; i++) {
      int r = i + (rand() % (size-i)); // Random remaining position.
      T temp = table[i]; table[i] = table[r]; table[r] = temp;
    }
  }
	
	private: bool isSorted(T* table,int size) const{
		bool sorted=true;
		
		for(int i = 1 ; sorted && i<size  ;i++)
			if(table[i]<table[i-1])
				sorted=false;
			
		return sorted;
	}
	
	//check if the algorithm sorts correctly
	public: void test(){
		int const size=7;
		T* table = new T[size];
		for(int i =0;i<size;i++)
		      table[i]=(rand()%3==0)?i:i+1;
		
		T reference_table[size];
		for(int i =0;i<size;i++)
		      reference_table[i]=table[i];		
    
		//shuffle the table
		shuffle(table,size);
		
		//sort the table
		SortAlgorithm<T>* sortAlgorithm;
		sortAlgorithm=getSortAlgorithm(table,size);
		sortAlgorithm->sort();
		
		cout<<endl<<"Sorting with: "<<sortAlgorithm->getName()<<endl;
		
		//compare sorted table with reference table
		for(int i =0;i<size;i++){
		      if(table[i]==reference_table[i])
			  cout<<"Element : "<<table[i]<<" ok"<<endl;
		      else
			  cout<<table[i]<<" is not the same as "<<reference_table[i]<<endl;
		}

		 if(!isSorted(table,size))
			cout<<"Sorting  "<<sortAlgorithm->getName()<<" FAILED!!!!!";
		else
			cout<<"Sorting "<<sortAlgorithm->getName()<<" succeeded!";
		cout<<endl;
		
		//cleanup:
		delete sortAlgorithm;
		delete[] table;
	}
	
	private: void sortAndPrint(Stopwatch* watch,SortAlgorithm<T>* sa){
		watch->start();
		sa->sort();
		watch->stop();
		cout<<setw(17)<<watch->time();
	}
	
	private: void printHeader(){
		T* tbl = new T[1];
		SortAlgorithm<T>* sa;
		sa = getSortAlgorithm(tbl,1);
		cout<<endl<<endl<<"Measuring sorting algorithm: "<<sa->getName()<<endl;
		delete sa;
		delete[] tbl;
	}
	
	//measures the effectiveness of a sort algorithm
	public: void measure(){
		measure(128,1000000,1.8);
	}	
	public: void measure(int min, int max , double step){
		srand(time(0)); 
		int current=min;
		Stopwatch watch;
		
		printHeader();
		
		cout<<setw(17)<<" ";
		cout<<setw(17)<<"Random";
		cout<<setw(17)<<"Sorted";
		cout<<setw(17)<<"Reversed"<<endl;
		
		while(current<=max){
			T* table = new T[current];
			SortAlgorithm<T>* sortAlgorithm;
			sortAlgorithm=getSortAlgorithm(table,current);
			
			cout<<setw(17)<<current;			
			
			 //(1) sort a random table
			 for(int i=0;i<current;i++)
			     table[i]=rand(); 
			 
			 sortAndPrint(&watch,sortAlgorithm);   
			 
			 //(2) sort a sorted table
			 sortAndPrint(&watch,sortAlgorithm);     
			 
			 //(3) sort a descending sorted table
			 reverse(table,current);
			 sortAndPrint(&watch,sortAlgorithm);   
			
			 cout<<endl;		
			 current= (int) (current * step);
			 delete sortAlgorithm;
			 delete[] table;
		 }
	}
	
	
	//returns the SortAlgorithm object based on the enum	
	private: SortAlgorithm<T>* getSortAlgorithm(T* table, int size){
		SortAlgorithm<T> * sortAlgorithm;
		switch(algorithm){
			case insertion_sort:
				sortAlgorithm = new InsertionSort<T>(table,size);
				break;
			case selection_sort:
				sortAlgorithm = new SelectionSort<T>(table,size);
				break;
			case heap_sort:
				sortAlgorithm = new HeapSort<T>(table,size);
				break;
			case quick_sort:
				sortAlgorithm = new QuickSort<T>(table,size);
				break;
			case shell_sort_sedgewick:
				sortAlgorithm = new ShellSort<T>(table,size,sedgewick);
				break;
			case shell_sort_two_increment:
				sortAlgorithm = new ShellSort<T>(table,size,two_increment);
				break;
			case shell_sort_product_of_three:
				sortAlgorithm = new ShellSort<T>(table,size,product_of_three);
				break;
      case quicker_sort:
				sortAlgorithm = new QuickerSort<T>(table,size,3);
				break;
      case six_sort:
				sortAlgorithm = new SixSort<T>(table,size);
				break;
      case counting_sort:
				sortAlgorithm = new CountingSort<T>(table,size);
				break;
      case quick_smart_pivot_sort:
        sortAlgorithm = new QuickSmartPivotSort<T>(table,size);
        break;
		}
		return sortAlgorithm;
	}
};

